class Point():
    y: float = None
    z: float = None
    def __init__(self, y: float, z: float) -> None:
        self.y = y
        self.z = z
    def __repr__(self) -> str:
        return f'<Point: {self.y:}, {self.z:}>'



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from .. import config

class TSection(GeneralSection):
    hw: float = None
    tw: float = None
    wf: float = None
    tf: float = None
    rf: float = None
    def __init__(self, hw: float, tw: float, wf: float, tf: float,
                 rf: float=0.0, label: str=None):
        self.hw = hw
        self.tw = tw
        self.wf = wf
        self.tf = tf
        self.rf = rf
        y = [-self.wf/2, -self.wf/2, -self.tw/2, -self.tw/2,
             self.tw/2, self.tw/2, self.wf/2, self.wf/2]
        z = [self.hw, self.hw-self.tf, self.hw-self.tf, 0.0,
             0.0, self.hw-self.tf, self.hw-self.tf, self.hw]
        r = [0.0, 0.0, self.rf, 0.0, 0.0, self.rf, 0.0, 0.0]
        super().__init__(y, z, r, label=label)
    def _repr_markdown_(self) -> str:
        mdstr = self.section_heading('T-Section')
        table = MDTable()
        table.add_column(f'h<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.hw])
        table.add_column(f't<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tw])
        table.add_column(f'w<sub>f</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.wf])
        table.add_column(f't<sub>f</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tf])
        table.add_column(f'r<sub>f</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.rf])
        mdstr += table._repr_markdown_()
        mdstr += self.section_properties(outtype='md')
        return mdstr
    def __str__(self) -> str:
        mdstr = self.section_heading('T-Section')
        table = MDTable()
        table.add_column(f'h_w ({config.lunit:s})', config.l1frm, data=[self.hw])
        table.add_column(f't_w ({config.lunit:s})', config.l1frm, data=[self.tw])
        table.add_column(f'w_f ({config.lunit:s})', config.l1frm, data=[self.wf])
        table.add_column(f't_f ({config.lunit:s})', config.l1frm, data=[self.tf])
        table.add_column(f'r_f ({config.lunit:s})', config.l1frm, data=[self.rf])
        mdstr += str(table)
        mdstr += self.section_properties(outtype='str')
        return mdstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<T-Section>'
        else:
            outstr = f'<T-Section {self.label:s}>'
        return outstr



from .lsectionformed import LSectionFormed as LSectionFormed
from .csectionformed import CSectionFormed as CSectionFormed
from .zsectionformed import ZSectionFormed as ZSectionFormed
from .omegasectionformed import OmegaSectionFormed as OmegaSectionFormed



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from ..general.thinwalledsection import ThinWalledSection
from .. import config

class ZSection(GeneralSection):
    hw: float = None
    tw: float = None
    wlf: float = None
    tlf: float = None
    wuf: float = None
    tuf: float = None
    rlf: float = None
    ruf: float = None
    def __init__(self, hw: float, tw: float, wlf: float, tlf: float,
                 wuf: float, tuf: float, ruf: float=0.0, rlf: float=0.0,
                 label: str=None) -> None:
        self.hw = hw
        self.tw = tw
        self.wlf = wlf
        self.tlf = tlf
        self.wuf = wuf
        self.tuf = tuf
        self.ruf = ruf
        self.rlf = rlf
        y = [0.0, self.wlf, self.wlf, self.tw, self.tw,
             self.tw-self.wuf, self.tw-self.wuf, 0.0]
        z = [0.0, 0.0, self.tlf, self.tlf, self.hw,
             self.hw, self.hw - self.tuf, self.hw - self.tuf]
        r = [0.0, 0.0, 0.0, self.ruf, 0.0, 0.0, 0.0, self.rlf]
        super().__init__(y, z, r, label=label)
    def to_thin_walled_section(self) -> ThinWalledSection:
        y = [self.wlf, self.tw/2, self.tw/2, self.tw/2-self.wuf]
        z = [self.tlf/2, self.tlf/2, self.hw-self.tuf/2, self.hw-self.tuf/2]
        t = [self.tlf, self.tw, self.tuf]
        return ThinWalledSection(y, z, t, label=self.label)
    def _repr_markdown_(self) -> str:
        mdstr = self.section_heading('Z-Section')
        table = MDTable()
        table.add_column(f'h<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.hw])
        table.add_column(f't<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tw])
        table.add_column(f'w<sub>lf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.wlf])
        table.add_column(f't<sub>lf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tlf])
        table.add_column(f'w<sub>uf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.wuf])
        table.add_column(f't<sub>uf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tuf])
        table.add_column(f'r<sub>uf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.ruf])
        table.add_column(f'r<sub>lf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.rlf])
        mdstr += table._repr_markdown_()
        mdstr += self.section_properties(outtype='md')
        return mdstr
    def __str__(self) -> str:
        outstr = self.section_heading('Z-Section')
        table = MDTable()
        table.add_column(f'h_w ({config.lunit:s})', config.l1frm, data=[self.hw])
        table.add_column(f't_w ({config.lunit:s})', config.l1frm, data=[self.tw])
        table.add_column(f'w_uf ({config.lunit:s})', config.l1frm, data=[self.wlf])
        table.add_column(f't_uf ({config.lunit:s})', config.l1frm, data=[self.tlf])
        table.add_column(f'w_lf ({config.lunit:s})', config.l1frm, data=[self.wuf])
        table.add_column(f't_lf ({config.lunit:s})', config.l1frm, data=[self.tuf])
        table.add_column(f'r_uf ({config.lunit:s})', config.l1frm, data=[self.ruf])
        table.add_column(f'r_lf ({config.lunit:s})', config.l1frm, data=[self.rlf])
        outstr += str(table)
        outstr += self.section_properties(outtype='str')
        return outstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<Z-Section>'
        else:
            outstr = f'<Z-Section {self.label:s}>'
        return outstr



from typing import List, Tuple, Optional
from math import sqrt
from py2md.classes import MDTable
from .point import Point
from .. import config

class Line():
    pnta: Point = None
    pntb: Point = None
    _length: float = None
    _A: float = None
    _Ay: float = None
    _Az: float = None
    _Ayy: float = None
    _Azz: float = None
    _Ayz: float = None
    def __init__(self, pnta: Point, pntb: Point):
        self.pnta = pnta
        self.pntb = pntb
    @property
    def length(self) -> float:
        if self._length is None:
            dy = self.pntb.y-self.pnta.y
            dz = self.pntb.z-self.pnta.z
            self._length = sqrt(dy**2+dz**2)
        return self._length
    @property
    def A(self) -> float:
        if self._A is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._A = (ya*zb-za*yb)/2
        return self._A
    @property
    def Ay(self) -> float:
        if self._Ay is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Ay = (ya*zb-za*yb)*(ya+yb)/6
        return self._Ay
    @property
    def Az(self) -> float:
        if self._Az is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Az = (ya*zb-za*yb)*(za+zb)/6
        return self._Az
    @property
    def Ayy(self) -> float:
        if self._Ayy is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Ayy = (ya**2+ya*yb+yb**2)*(ya*zb-yb*za)/12
        return self._Ayy
    @property
    def Azz(self) -> float:
        if self._Azz is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Azz = (za**2+za*zb+zb**2)*(ya*zb-yb*za)/12
        return self._Azz
    @property
    def Ayz(self) -> float:
        if self._Ayz is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Ayz = (ya*zb+2*ya*za+2*yb*zb+yb*za)*(ya*zb-yb*za)/24
        return self._Ayz
    def add_path(self, verts: Optional[List[Tuple[float, float]]]=None,
                 codes: Optional[List[float]]=None) -> Tuple[List[Tuple[float, float]],
                                                             List[float]]:
        if verts is None:
            verts = []
        if codes is None:
            codes = []
        if len(verts) == 0:
            verts.append((self.pnta.y, self.pnta.z))
            codes.append(1)
        verts.append((self.pntb.y, self.pntb.z))
        codes.append(2)
        return verts, codes
    def _repr_markdown_(self) -> str:
        table = MDTable()
        table.add_column('Point', 's')
        table.add_column('y', config.l1frm)
        table.add_column('z', config.l1frm)
        table.add_row(['pnta', self.pnta.y, self.pnta.z])
        table.add_row(['pntb', self.pntb.y, self.pntb.z])
        mdstr = table._repr_markdown_()
        table = MDTable()
        table.add_column('Type', 's')
        table.add_column('A', config.l2frm)
        table.add_column('Ay', config.l3frm)
        table.add_column('Az', config.l3frm)
        table.add_column('Ayy', config.l4frm)
        table.add_column('Azz', config.l4frm)
        table.add_column('Ayz', config.l4frm)
        table.add_row(['Total', self.A, self.Ay, self.Az,
                       self.Ayy, self.Azz, self.Ayz])
        mdstr += table._repr_markdown_()
        return mdstr
    def __str__(self) -> str:
        table = MDTable()
        table.add_column('Point', 's')
        table.add_column('y', config.l1frm)
        table.add_column('z', config.l1frm)
        table.add_row(['pnta', self.pnta.y, self.pnta.z])
        table.add_row(['pntb', self.pntb.y, self.pntb.z])
        outstr = str(table)
        table = MDTable()
        table.add_column('Type', 's')
        table.add_column('A', config.l2frm)
        table.add_column('Ay', config.l3frm)
        table.add_column('Az', config.l3frm)
        table.add_column('Ayy', config.l4frm)
        table.add_column('Azz', config.l4frm)
        table.add_column('Ayz', config.l4frm)
        table.add_row(['Total', self.A, self.Ay, self.Az,
                       self.Ayy, self.Azz, self.Ayz])
        outstr += str(table)
        return outstr
    def __repr__(self) -> str:
        return f'<Line: {self.pnta}, {self.pntb}>'



from typing import Optional
from py2md.classes import MDTable, MDHeading
from .. import config

class Material():
    E: float = None
    Ec: float = None
    Ftu: float = None
    Fcu: float = None
    Fty: float = None
    Fcy: float = None
    label: str = None
    def __init__(self, E: float, Ec: float=None, label: str=None) -> None:
        self.E = E
        self.Ec = Ec
        if self.Ec is None:
            self.Ec = self.E
        self.label = label
    def set_yield_strengths(self, Fty: float, Fcy: float) -> None:
        self.Fty = Fty
        self.Fcy = Fcy
    def set_ultimate_strengths(self, Ftu: float, Fcu: Optional[float]=None) -> None:
        self.Ftu = Ftu
        if Fcu is None:
            self.Fcu = Ftu
        else:
            self.Fcu = Fcu
    def _repr_markdown_(self) -> str:
        if self.label is None:
            heading  = MDHeading('Material Properties', 3)
        else:
            heading  = MDHeading(f'Material Properties - {self.label:s}', 3)
        mdstr = str(heading)
        table = MDTable()
        table.add_column(f'E ({config.sunit:s})', config.efrm, data=[self.E])
        table.add_column(f'E<sub>c</sub> ({config.sunit:s})',
                         config.efrm, data=[self.Ec])
        if self.Fty is not None:
            table.add_column(f'F<sub>ty</sub> ({config.sunit:s})',
                             config.sfrm, data=[self.Fty])
        if self.Fcy is not None:
            table.add_column(f'F<sub>cy</sub> ({config.sunit:s})',
                             config.sfrm, data=[self.Fcy])
        if self.Ftu is not None:
            table.add_column(f'F<sub>tu</sub> ({config.sunit:s})',
                             config.sfrm, data=[self.Ftu])
        if self.Fcu is not None:
            table.add_column(f'F<sub>cu</sub> ({config.sunit:s})',
                             config.sfrm, data=[self.Fcu])
        mdstr += table._repr_markdown_()
        return mdstr
    def __str__(self) -> str:
        if self.label is None:
            heading  = MDHeading('Material Properties', 3)
        else:
            heading  = MDHeading(f'Material Properties - {self.label:s}', 3)
        outstr = str(heading)
        table = MDTable()
        table.add_column(f'E ({config.sunit:s})', config.efrm, data=[self.E])
        table.add_column(f'Ec ({config.sunit:s})',
                         config.efrm, data=[self.Ec])
        if self.Fty is not None:
            table.add_column(f'Fty ({config.sunit:s})',
                             config.sfrm, data=[self.Fty])
        if self.Fcy is not None:
            table.add_column(f'Fcy ({config.sunit:s})',
                             config.sfrm, data=[self.Fcy])
        if self.Ftu is not None:
            table.add_column(f'Ftu ({config.sunit:s})',
                             config.sfrm, data=[self.Ftu])
        if self.Fcu is not None:
            table.add_column(f'Fcu ({config.sunit:s})',
                             config.sfrm, data=[self.Fcu])
        outstr += str(table)
        return outstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<Material>'
        else:
            outstr = f'<Material {self.label:s}>'
        return outstr



from .general.generalsection import GeneralSection as GeneralSection
from .general.materialsection import MaterialSection as MaterialSection
from .general.compositesection import CompositeSection as CompositeSection
from .general.thinwalledsection import ThinWalledSection as ThinWalledSection
from .general.cripplingsection import CripplingSection as CripplingSection



lunit = 'mm' # Length unit used in pysectprop
funit = 'N' # Force unit used in pysectprop
sunit = 'MPa' # Stress unit used in pysectprop

angfrm = '.1f' # Angle format
l1frm = '.1f' # Length format
l2frm = '.1f' # Length**2 format
l3frm = '.0f' # Length**3 format
l4frm = '.0f' # Length**4 format
sfrm = '.1f' # Stress format
efrm = '.0f' # Elastic Modulus formaty

msmode = False # Output Margins of Safety



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from ..general.thinwalledsection import ThinWalledSection
from .. import config

class OmegaSectionFormed(GeneralSection):
    hw: float = None # Height of web
    wuf: float = None # Width of lower flange
    wlf: float = None # Width of upper flange
    ts: float = None # Thickness of section
    rm: float = None # Bend radius
    def __init__(self, hw: float, wlf: float, wuf: float, ts: float,
                 rm: float, label: str=None) -> None:
        self.hw = hw
        self.wlf = wlf
        self.wuf = wuf
        self.ts = ts
        self.rm = rm
        y = [-self.wlf-self.wuf/2+self.ts, -self.wuf/2+self.ts,
             -self.wuf/2+self.ts, self.wuf/2-self.ts, self.wuf/2-self.ts,
             self.wlf+self.wuf/2-self.ts, self.wlf+self.wuf/2-self.ts,
             self.wuf/2, self.wuf/2, -self.wuf/2, -self.wuf/2,
             -self.wlf-self.wuf/2+self.ts]
        z = [0.0, 0.0, self.hw-self.ts, self.hw-self.ts, 0.0, 0.0,
             self.ts, self.ts, self.hw, self.hw, self.ts, self.ts]
        r = [0.0, self.rm+self.ts/2, self.rm-self.ts/2, self.rm-self.ts/2,
             self.rm+self.ts/2, 0.0, 0.0, self.rm-self.ts/2, self.rm+self.ts/2,
             self.rm+self.ts/2, self.rm-self.ts/2, 0.0]
        super().__init__(y, z, r, label=label)
    def to_thin_walled_section(self) -> ThinWalledSection:
        y = [-self.wlf-self.wuf/2+self.ts, -self.wuf/2+self.ts/2, -self.wuf/2+self.ts/2,
             self.wuf/2-self.ts/2, self.wuf/2-self.ts/2, self.wlf+self.wuf/2-self.ts]
        z = [self.ts/2, self.ts/2, self.hw-self.ts/2, self.hw-self.ts/2,
             self.ts/2, self.ts/2]
        t = [self.ts, self.ts, self.ts, self.ts, self.ts]
        return ThinWalledSection(y, z, t)
    def _repr_markdown_(self) -> str:
        mdstr = self.section_heading('Omega-Section Formed')
        table = MDTable()
        table.add_column(f'h<sub>w</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.hw])
        table.add_column(f'w<sub>lf</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.wlf])
        table.add_column(f'w<sub>uf</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.wuf])
        table.add_column(f't<sub>s</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.ts])
        table.add_column(f'r<sub>m</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.rm])
        mdstr += table._repr_markdown_()
        mdstr += self.section_properties(outtype='md')
        return mdstr
    def __str__(self) -> str:
        outstr = self.section_heading('Omega-Section Formed')
        table = MDTable()
        table.add_column(f'h<sub>w</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.hw])
        table.add_column(f'w<sub>lf</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.wlf])
        table.add_column(f'w<sub>uf</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.wuf])
        table.add_column(f't<sub>s</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.ts])
        table.add_column(f'r<sub>m</sub> ({config.lunit:s})',
                         config.l1frm, data=[self.rm])
        outstr += str(table)
        outstr += self.section_properties(outtype='str')
        return outstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<Omega-Section Formed>'
        else:
            outstr = f'<Omega-Section Formed: {self.label:s}>'
        return outstr



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from .. import config

class LSection(GeneralSection):
    hw: float = None
    tw: float = None
    wf: float = None
    tf: float = None
    rc: float = None
    def __init__(self, hw: float, tw: float, wf: float, tf: float,
                 rc: float=0.0, label: str=None) -> None:
        self.hw = hw
        self.tw = tw
        self.wf = wf
        self.tf = tf
        self.rc = rc
        y = [0.0, self.wf, self.wf, self.tw, self.tw, 0.0]
        z = [0.0, 0.0, self.tf, self.tf, self.hw, self.hw]
        r = [0.0, 0.0, 0.0, self.rc, 0.0, 0.0]
        super().__init__(y, z, r, label=label)
    def _repr_markdown_(self) -> str:
        mdstr = self.section_heading('L-Section')
        table = MDTable()
        table.add_column(f'h<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.hw])
        table.add_column(f't<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tw])
        table.add_column(f'w<sub>f</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.wf])
        table.add_column(f't<sub>f</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tf])
        table.add_column(f'r<sub>c</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.rc])
        mdstr += table._repr_markdown_()
        mdstr += self.section_properties(outtype='md')
        return mdstr
    def __str__(self) -> str:
        outstr = self.section_heading('L-Section')
        table = MDTable()
        table.add_column(f'h_w ({config.lunit:s})', config.l1frm, data=[self.hw])
        table.add_column(f't_w ({config.lunit:s})', config.l1frm, data=[self.tw])
        table.add_column(f'w_f ({config.lunit:s})', config.l1frm, data=[self.wf])
        table.add_column(f't_f ({config.lunit:s})', config.l1frm, data=[self.tf])
        table.add_column(f'r_c ({config.lunit:s})', config.l1frm, data=[self.rc])
        outstr += str(table)
        outstr += self.section_properties(outtype='str')
        return outstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<L-Section>'
        else:
            outstr = f'<L-Section {self.label:s}>'
        return outstr



from .generalsection import GeneralSection as GeneralSection
from .materialsection import MaterialSection as MaterialSection
from .compositesection import CompositeSection as CompositeSection
from .cripplingsection import CripplingSection as CripplingSection
from .thinwalledsection import ThinWalledSection as ThinWalledSection
from .material import Material as Material



from typing import List, Tuple, Optional
from math import atan2, sqrt
from py2md.classes import MDTable
from .point import Point
from .. import config

K = 1.0 - 4.0*(2.0**0.5 - 1.0)/3.0

class Arc():
    pnta: Point = None
    pntb: Point = None
    pntc: Point = None
    pntd: Point = None
    pnte: Point = None
    pntf: Point = None
    sinang: float = None
    cosang: float = None
    ang: float = None
    sango2: float = None
    radius: float = None
    sinphi: float = None
    cosphi: float = None
    _A: float = None
    _Ay: float = None
    _Az: float = None
    _Ayy: float = None
    _Azz: float = None
    _Ayz: float = None
    _A_seg: float = None
    _Ay_seg: float = None
    _Az_seg: float = None
    _Ayy_seg: float = None
    _Azz_seg: float = None
    _Ayz_seg: float = None
    _A_line: float = None
    _Ay_line: float = None
    _Az_line: float = None
    _Ayy_line: float = None
    _Azz_line: float = None
    _Ayz_line: float = None
    def __init__(self, pnta: Point, pntb: Point, pntc: Point, pntf: Point) -> None:
        self.pnta = pnta
        self.pntb = pntb
        self.pntc = pntc
        self.pntf = pntf
        self.update()
    def update(self) -> None:
        dy = self.pntb.y - self.pnta.y
        dz = self.pntb.z - self.pnta.z
        length = sqrt(dy**2 + dz**2)
        dy1 = self.pnta.y - self.pntc.y
        dz1 = self.pnta.z - self.pntc.z
        dy2 = self.pntb.y - self.pntc.y
        dz2 = self.pntb.z - self.pntc.z
        yd = self.pntc.y + dy1*K
        zd = self.pntc.z + dz1*K
        ye = self.pntc.y + dy2*K
        ze = self.pntc.z + dz2*K
        self.pntd = Point(yd, zd)
        self.pnte = Point(ye, ze)
        l1 = sqrt(dy1**2 + dz1**2)
        dy3 = self.pnta.y - self.pntf.y
        dz3 = self.pnta.z - self.pntf.z
        dy4 = self.pntb.y - self.pntf.y
        dz4 = self.pntb.z - self.pntf.z
        l3 = sqrt(dy3**2 + dz3**2)
        l4 = sqrt(dy4**2 + dz4**2)
        self.sinang = (dy3*dz4 - dz3*dy4)/l3/l4
        self.cosang = (dy3*dy4 + dz3*dz4)/l3/l4
        self.ang = atan2(self.sinang, self.cosang)
        self.sango2 = (dy*dz1 - dz*dy1)/length/l1
        yaf = self.pnta.y - self.pntf.y
        zaf = self.pnta.z - self.pntf.z
        self.radius = sqrt(yaf**2 + zaf**2)
        zcf = self.pntc.z - self.pntf.z
        ycf = self.pntc.y - self.pntf.y
        lcf = sqrt(ycf**2 + zcf**2)
        self.sinphi = zcf/lcf
        self.cosphi = ycf/lcf
    @property
    def A_seg(self) -> float:
        if self._A_seg is None:
            rad = self.radius
            ang = self.ang
            sinang = self.sinang
            self._A_seg = rad**2/2*(ang-sinang)
        return self._A_seg
    @property
    def A_line(self) -> float:
        if self._A_line is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._A_line = (ya*zb-za*yb)/2
        return self._A_line
    @property
    def A(self) -> float:
        if self._A is None:
            self._A = self.A_seg + self.A_line
        return self._A
    @property
    def Ay_seg(self) -> float:
        if self._Ay_seg is None:
            yf = self.pntf.y
            rad = self.radius
            sango2 = self.sango2
            cosphi = self.cosphi
            self._Ay_seg = 2/3*rad**3*sango2**3*cosphi+self.A_seg*yf
        return self._Ay_seg
    @property
    def Ay_line(self) -> float:
        if self._Ay_line is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Ay_line = (ya*zb-za*yb)*(ya+yb)/6
        return self._Ay_line
    @property
    def Ay(self) -> float:
        if self._Ay is None:
            self._Ay = self.Ay_seg+self.Ay_line
        return self._Ay
    @property
    def Az_seg(self) -> float:
        if self._Az_seg is None:
            zf = self.pntf.z
            rad = self.radius
            sango2 = self.sango2
            sinphi = self.sinphi
            self._Az_seg = 2/3*rad**3*sango2**3*sinphi+self._A_seg*zf
        return self._Az_seg
    @property
    def Az_line(self) -> float:
        if self._Az_line is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Az_line = (ya*zb-za*yb)*(za+zb)/6
        return self._Az_line
    @property
    def Az(self) -> float:
        if self._Az is None:
            self._Az = self.Az_seg+self.Az_line
        return self._Az
    @property
    def Ayy_seg(self) -> float:
        if self._Ayy_seg is None:
            yf = self.pntf.y
            rad = self.radius
            ang = self.ang
            sango2 = self.sango2
            sinang = self.sinang
            sinphi = self.sinphi
            cosphi = self.cosphi
            Ayy = rad**4/8*(ang-sinang+2*sinang*sango2**2)
            Azz = rad**4/8*(ang-sinang-2*sinang*sango2**2/3)
            self._Ayy_seg = Ayy*cosphi**2+Azz*sinphi**2+self.A_seg*yf**2
        return self._Ayy_seg
    @property
    def Ayy_line(self) -> float:
        if self._Ayy_line is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Ayy_line = (ya**2+ya*yb+yb**2)*(ya*zb-yb*za)/12
        return self._Ayy_line
    @property
    def Ayy(self) -> float:
        if self._Ayy is None:
            self._Ayy = self.Ayy_seg+self.Ayy_line
        return self._Ayy
    @property
    def Azz_seg(self) -> float:
        if self._Azz_seg is None:
            zf = self.pntf.z
            rad = self.radius
            ang = self.ang
            sango2 = self.sango2
            sinang = self.sinang
            sinphi = self.sinphi
            cosphi = self.cosphi
            Ayy = rad**4/8*(ang-sinang+2*sinang*sango2**2)
            Azz = rad**4/8*(ang-sinang-2*sinang*sango2**2/3)
            self._Azz_seg = Azz*cosphi**2+Ayy*sinphi**2+self.A_seg*zf**2
        return self._Azz_seg
    @property
    def Azz_line(self) -> float:
        if self._Azz_line is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Azz_line = (za**2+za*zb+zb**2)*(ya*zb-yb*za)/12
        return self._Azz_line
    @property
    def Azz(self) -> float:
        if self._Azz is None:
            self._Azz = self.Azz_seg+self.Azz_line
        return self._Azz
    @property
    def Ayz_seg(self) -> float:
        if self._Ayz_seg is None:
            yf = self.pntf.y
            zf = self.pntf.z
            rad = self.radius
            ang = self.ang
            sango2 = self.sango2
            sinang = self.sinang
            sinphi = self.sinphi
            cosphi = self.cosphi
            A_seg = rad**2/2*(ang-sinang)
            tempAyy = rad**4/8*(ang-sinang+2*sinang*sango2**2)
            tempAzz = rad**4/8*(ang-sinang-2*sinang*sango2**2/3)
            self._Ayz_seg = (tempAyy-tempAzz)*sinphi*cosphi+A_seg*yf*zf
        return self._Ayz_seg
    @property
    def Ayz_line(self) -> float:
        if self._Ayz_line is None:
            ya = self.pnta.y
            za = self.pnta.z
            yb = self.pntb.y
            zb = self.pntb.z
            self._Ayz_line = (ya*zb+2*ya*za+2*yb*zb+yb*za)*(ya*zb-yb*za)/24
        return self._Ayz_line
    @property
    def Ayz(self) -> float:
        if self._Ayz is None:
            self._Ayz = self.Ayz_seg + self.Ayz_line
        return self._Ayz
    def add_path(self, verts: Optional[List[Tuple[float, float]]]=None,
                 codes: Optional[List[float]]=None) -> Tuple[List[Tuple[float, float]],
                                                             List[float]]:
        if verts is None:
            verts = []
        if codes is None:
            codes = []
        if len(verts) == 0:
            verts.append((self.pnta.y, self.pnta.z))
            codes.append(1)
        verts.append((self.pntd.y, self.pntd.z))
        codes.append(4)
        verts.append((self.pnte.y, self.pnte.z))
        codes.append(4)
        verts.append((self.pntb.y, self.pntb.z))
        codes.append(4)
        return verts, codes
    def _repr_markdown_(self) -> str:
        table = MDTable()
        table.add_column('Point', 's')
        table.add_column('y', config.l1frm)
        table.add_column('z', config.l1frm)
        table.add_row(['pnta', self.pnta.y, self.pnta.z])
        table.add_row(['pntb', self.pntb.y, self.pntb.z])
        table.add_row(['pntc', self.pntc.y, self.pntc.z])
        table.add_row(['pntd', self.pntd.y, self.pntd.z])
        table.add_row(['pnte', self.pnte.y, self.pnte.z])
        table.add_row(['pntf', self.pntf.y, self.pntf.z])
        mdstr = table._repr_markdown_()
        table = MDTable()
        table.add_column('Type', 's')
        table.add_column('A', config.l2frm)
        table.add_column('Ay', config.l3frm)
        table.add_column('Az', config.l3frm)
        table.add_column('Ayy', config.l4frm)
        table.add_column('Azz', config.l4frm)
        table.add_column('Ayz', config.l4frm)
        table.add_row(['Segment', self.A_seg, self.Ay_seg, self.Az_seg,
                       self.Ayy_seg, self.Azz_seg, self.Ayz_seg])
        table.add_row(['Line', self.A_line, self.Ay_line, self.Az_line,
                       self.Ayy_line, self.Azz_line, self.Ayz_line])
        table.add_row(['Total', self.A, self.Ay, self.Az,
                       self.Ayy, self.Azz, self.Ayz])
        mdstr += table._repr_markdown_()
        return mdstr
    def __str__(self) -> str:
        table = MDTable()
        table.add_column('Point', 's')
        table.add_column('y', config.l1frm)
        table.add_column('z', config.l1frm)
        table.add_row(['pnta', self.pnta.y, self.pnta.z])
        table.add_row(['pntb', self.pntb.y, self.pntb.z])
        table.add_row(['pntc', self.pntc.y, self.pntc.z])
        table.add_row(['pntd', self.pntd.y, self.pntd.z])
        table.add_row(['pnte', self.pnte.y, self.pnte.z])
        table.add_row(['pntf', self.pntf.y, self.pntf.z])
        outstr = str(table)
        table = MDTable()
        table.add_column('Type', 's')
        table.add_column('A', config.l2frm)
        table.add_column('Ay', config.l3frm)
        table.add_column('Az', config.l3frm)
        table.add_column('Ayy', config.l4frm)
        table.add_column('Azz', config.l4frm)
        table.add_column('Ayz', config.l4frm)
        table.add_row(['Segment', self.A_seg, self.Ay_seg, self.Az_seg,
                       self.Ayy_seg, self.Azz_seg, self.Ayz_seg])
        table.add_row(['Line', self.A_line, self.Ay_line, self.Az_line,
                       self.Ayy_line, self.Azz_line, self.Ayz_line])
        table.add_row(['Total', self.A, self.Ay, self.Az,
                       self.Ayy, self.Azz, self.Ayz])
        outstr += str(table)
        return outstr
    def __repr__(self) -> str:
        return f'<Arc: {self.pnta}, {self.pntc}, {self.pntb}>'

def arc_from_points(pnta: Point, pntb: Point, pntc: Point, radius: float) -> Arc:
    dy1 = pnta.y - pntb.y
    dz1 = pnta.z - pntb.z
    dy2 = pntc.y - pntb.y
    dz2 = pntc.z - pntb.z
    l1 = sqrt(dy1**2 + dz1**2)
    l2 = sqrt(dy2**2 + dz2**2)
    dy1 = dy1/l1
    dz1 = dz1/l1
    dy2 = dy2/l2
    dz2 = dz2/l2
    sinab = dy2*dz1 - dz2*dy1
    cosab = dy1*dy2 + dz1*dz2
    cotabo2 = (1 + cosab)/sinab
    lp = abs(radius*cotabo2)
    yd = pntb.y + dy1*lp
    zd = pntb.z + dz1*lp
    ye = pntb.y + dy2*lp
    ze = pntb.z + dz2*lp
    pntf = determine_pntf(yd, zd, -dz1, dy1, ye, ze, -dz2, dy2)
    pntd = Point(yd, zd)
    pnte = Point(ye, ze)
    arc = Arc(pntd, pnte, pntb, pntf)
    return arc

def determine_pntf(ya, za, va, wa, yb, zb, vb, wb) -> Point:
    lbres = (va*za - va*zb - wa*ya + wa*yb)/(va*wb - vb*wa)
    yf = vb*lbres + yb
    zf = wb*lbres + zb
    return Point(yf, zf)



from .circlesection import CircleSection as CircleSection
from .rectanglesection import RectangleSection as RectangleSection
from .tubesection import TubeSection as TubeSection
from .semitubesection import SemiTubeSection as SemiTubeSection


from typing import TYPE_CHECKING
from math import degrees
from py2md.classes import MDHeading, MDTable
from .generalsection import GeneralSection
from ..results.sectionresult import SectionResult
from .. import config

if TYPE_CHECKING:
    from .material import Material

class MaterialSection(GeneralSection):
    material: 'Material' = None
    _EA: float = None
    _EAy: float = None
    _EAz: float = None
    _EAyy: float = None
    _EAzz: float = None
    _EAyz: float = None
    _EIyy: float = None
    _EIzz: float = None
    _EIyz: float = None
    _EIyp: float = None
    _EIzp: float = None
    def __init__(self, section: GeneralSection, material: 'Material'):
        super().__init__(section.y, section.z, section.r, label=section.label)
        for k in section.__dict__:
            self.__dict__[k] = section.__dict__[k]
        self.material = material
    @property
    def EA(self) -> float:
        if self._EA is None:
            self._EA = self.material.E*self.A
        return self._EA
    @property
    def EAy(self) -> float:
        if self._EAy is None:
            self._EAy = self.material.E*self.Ay
        return self._EAy
    @property
    def EAz(self) -> float:
        if self._EAz is None:
            self._EAz = self.material.E*self.Az
        return self._EAz
    @property
    def EAyy(self) -> float:
        if self._EAyy is None:
            self._EAyy = self.material.E*self.Ayy
        return self._EAyy
    @property
    def EAzz(self) -> float:
        if self._EAzz is None:
            self._EAzz = self.material.E*self.Azz
        return self._EAzz
    @property
    def EAyz(self) -> float:
        if self._EAyz is None:
            self._EAyz = self.material.E*self.Ayz
        return self._EAyz
    @property
    def EIyy(self) -> float:
        if self._EIyy is None:
            self._EIyy = self.material.E*self.Iyy
        return self._EIyy
    @property
    def EIzz(self) -> float:
        if self._EIzz is None:
            self._EIzz = self.material.E*self.Izz
        return self._EIzz
    @property
    def EIyz(self) -> float:
        if self._EIyz is None:
            self._EIyz = self.material.E*self.Iyz
        return self._EIyz
    @property
    def EIyp(self) -> float:
        if self._EIyp is None:
            self._EIyp = self.material.E*self.Iyp
        return self._EIyp
    @property
    def EIzp(self) -> float:
        if self._EIzp is None:
            self._EIzp = self.material.E*self.Izp
        return self._EIzp
    def apply_load(self, loadcase: str, lctype: str,
                   Fx: float, My: float, Mz: float) -> SectionResult:
        sectresult = SectionResult(self)
        sectresult.set_load(loadcase, lctype, Fx, My, Mz)
        return sectresult
    def _repr_markdown_(self) -> str:
        funit = config.funit
        lunit = config.lunit
        l1frm = config.l1frm
        l2frm = config.l2frm
        l3frm = config.l3frm
        l4frm = config.l4frm
        angfrm = config.angfrm
        eiunit = f'{funit:s}.{lunit:s}<sup>2</sup>'
        mdstr = ''
        if self.label is None:
            head = 'Material Section Properties'
        else:
            head = f'Material Section Properties - {self.label:s}'
        heading = MDHeading(head, 3)
        mdstr += str(heading)
        table = MDTable()
        table.add_column(f'EA ({funit:s})', l2frm, data=[self.EA])
        table.add_column(f'EAy ({funit:s}.{lunit:s})', l3frm, data=[self.EAy])
        table.add_column(f'EAz ({funit:s}.{lunit:s})', l3frm, data=[self.EAz])
        table.add_column(f'cy ({lunit:s})', l1frm, data=[self.cy])
        table.add_column(f'cz ({lunit:s})', l1frm, data=[self.cz])
        table.add_column(f'EAyy ({eiunit:s})', l4frm, data=[self.EAyy])
        table.add_column(f'EAzz ({eiunit:s})', l4frm, data=[self.EAzz])
        table.add_column(f'EAyz ({eiunit:s})', l4frm, data=[self.EAyz])
        mdstr += table._repr_markdown_()
        table = MDTable()
        table.add_column(f'EI<sub>yy</sub> ({eiunit:s})', l4frm, data=[self.EIyy])
        table.add_column(f'EI<sub>zz</sub> ({eiunit:s})', l4frm, data=[self.EIzz])
        table.add_column(f'EI<sub>yz</sub> ({eiunit:s})', l4frm, data=[self.EIyz])
        table.add_column('&theta;<sub>p</sub> (&deg;)', angfrm,
                         data=[degrees(self.thp)])
        table.add_column(f'EI<sub>yp</sub> ({eiunit:s})', l4frm, data=[self.EIyp])
        table.add_column(f'EI<sub>zp</sub> ({eiunit:s})', l4frm, data=[self.EIzp])
        mdstr += table._repr_markdown_()
        return mdstr
    def __str__(self) -> str:
        funit = config.funit
        lunit = config.lunit
        l1frm = config.l1frm
        l2frm = config.l2frm
        l3frm = config.l3frm
        l4frm = config.l4frm
        angfrm = config.angfrm
        eiunit = f'{funit:s}.{lunit:s}^2'
        outstr = ''
        if self.label is None:
            head = 'Material Section Properties'
        else:
            head = f'Material Section Properties - {self.label:s}'
        heading = MDHeading(head, 3)
        outstr += str(heading)
        table = MDTable()
        table.add_column(f'EA ({funit:s})', l2frm, data=[self.EA])
        table.add_column(f'EAy ({funit:s}.{lunit:s})', l3frm, data=[self.EAy])
        table.add_column(f'EAz ({funit:s}.{lunit:s})', l3frm, data=[self.EAz])
        table.add_column(f'cy ({lunit:s})', l1frm, data=[self.cy])
        table.add_column(f'cz ({lunit:s})', l1frm, data=[self.cz])
        table.add_column(f'EAyy ({eiunit:s})', l4frm, data=[self.EAyy])
        table.add_column(f'EAzz ({eiunit:s})', l4frm, data=[self.EAzz])
        table.add_column(f'EAyz ({eiunit:s})', l4frm, data=[self.EAyz])
        outstr += str(table)
        table = MDTable()
        table.add_column(f'EI_yy ({eiunit:s})', l4frm, data=[self.EIyy])
        table.add_column(f'EI_zz ({eiunit:s})', l4frm, data=[self.EIzz])
        table.add_column(f'EI_yz ({eiunit:s})', l4frm, data=[self.EIyz])
        table.add_column('th_p (deg)', angfrm,
                         data=[degrees(self.thp)])
        table.add_column(f'EI_yp ({eiunit:s})', l4frm, data=[self.EIyp])
        table.add_column(f'EI_zp ({eiunit:s})', l4frm, data=[self.EIzp])
        outstr += str(table)
        return outstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<MaterialSection>'
        else:
            outstr = f'<MaterialSection {self.label:s}>'
        return outstr



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from .. import config

class RectangleSection(GeneralSection):
    h: float = None
    b: float = None
    rc: float = None
    def __init__(self, h: float, b: float, rc: float=0.0, label: str=None) -> None:
        self.h = h
        self.b = b
        self.rc = rc
        y = [self.b/2,
             self.b/2, -self.b/2,
             -self.b/2]
        z = [-self.h/2,
             self.h/2, self.h/2, -self.h/2]
        r = [self.rc, self.rc, self.rc, self.rc]
        super().__init__(y, z, r, label=label)
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<Rectangle-Section>'
        else:
            outstr = f'<Rectangle-Section {self.label:s}>'
        return outstr
    def __str__(self) -> str:
        mdstr = self.section_heading('Rectangle-Section')
        table = MDTable()
        table.add_column(f'h ({config.lunit:s})', config.l1frm, data=[self.h])
        table.add_column(f'b ({config.lunit:s})', config.l1frm, data=[self.b])
        mdstr += str(table)
        mdstr += self.section_properties()
        return mdstr
    def _repr_markdown_(self) -> str:
        return self.__str__()



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from .. import config

class JSection(GeneralSection):
    hw: float = None
    tw: float = None
    wuf: float = None
    wlf: float = None
    tuf: float = None
    tlf: float = None
    ruf: float = None
    rlf: float = None
    def __init__(self, hw: float, tw: float, wuf: float, tuf: float,
                 wlf: float, tlf: float, rlf: float=0.0, ruf: float=0.0,
                 label: str=None) -> None:
        self.hw = hw
        self.tw = tw
        self.wuf = wuf
        self.tuf = tuf
        self.wlf = wlf
        self.tlf = tlf
        self.rlf = rlf
        self.ruf = ruf
        y = [self.tw/2, self.tw/2, self.wuf/2, self.wuf/2,
             -self.wuf/2, -self.wuf/2, -self.tw/2, -self.tw/2,
             -self.wlf, -self.wlf]
        z = [0.0, self.hw-self.tuf, self.hw-self.tuf, self.hw,
             self.hw, self.hw-self.tuf, self.hw-self.tuf,
             self.tlf, self.tlf, 0.0]
        r = [0.0, self.ruf, 0.0, 0.0, 0.0, 0.0, self.ruf,
             self.rlf, 0.0, 0.0]
        super().__init__(y, z, r, label=label)
    def _repr_markdown_(self) -> str:
        mdstr = self.section_heading('J-Section')
        table = MDTable()
        table.add_column(f'h<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.hw])
        table.add_column(f't<sub>w</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tw])
        table.add_column(f'w<sub>uf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.wuf])
        table.add_column(f't<sub>uf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tuf])
        table.add_column(f'w<sub>lf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.wlf])
        table.add_column(f't<sub>lf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.tlf])
        table.add_column(f'r<sub>lf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.rlf])
        table.add_column(f'r<sub>uf</sub> ({config.lunit:s})', config.l1frm,
                         data=[self.ruf])
        mdstr += table._repr_markdown_()
        mdstr += self.section_properties(outtype='md')
        return mdstr
    def __str__(self) -> str:
        outstr = self.section_heading('J-Section')
        table = MDTable()
        table.add_column(f'h_w ({config.lunit:s})', config.l1frm, data=[self.hw])
        table.add_column(f't_w ({config.lunit:s})', config.l1frm, data=[self.tw])
        table.add_column(f'w_uf ({config.lunit:s})', config.l1frm, data=[self.wuf])
        table.add_column(f't_uf ({config.lunit:s})', config.l1frm, data=[self.tuf])
        table.add_column(f'w_lf ({config.lunit:s})', config.l1frm, data=[self.wlf])
        table.add_column(f't_lf ({config.lunit:s})', config.l1frm, data=[self.tlf])
        table.add_column(f'r_lf ({config.lunit:s})', config.l1frm, data=[self.rlf])
        table.add_column(f'r_uf ({config.lunit:s})', config.l1frm, data=[self.ruf])
        outstr += str(table)
        outstr += self.section_properties(outtype='str')
        return outstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<J-Section>'
        else:
            outstr = f'<J-Section {self.label:s}>'
        return outstr



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from .. import config

class CircleSection(GeneralSection):
    d = None
    def __init__(self, d: float, label: str=None):
        self.d = d
        radius = self.d/2
        y = [radius, -radius, -radius, radius]
        z = [radius, radius, -radius, -radius]
        r = [radius, radius, radius, radius]
        super().__init__(y, z, r, label=label)
    def __repr__(self):
        if self.label is None:
            outstr = '<Circle-Section>'
        else:
            outstr = f'<Circle-Section {self.label:s}>'
        return outstr
    def __str__(self):
        mdstr = self.section_heading('Circle-Section')
        table = MDTable()
        table.add_column(f'Diameter ({config.lunit:s})', config.l1frm, data=[self.d])
        mdstr += str(table)
        mdstr += self.section_properties()
        return mdstr
    def _repr_markdown_(self):
        return self.__str__()



from .lsection import LSection as LSection
from .omegasection import OmegaSection as OmegaSection
from .isection import ISection as ISection
from .jsection import JSection as JSection
from .tsection import TSection as TSection
from .zsection import ZSection as ZSection



from typing import TYPE_CHECKING, List
from math import cos, sin, pi, atan, degrees
from matplotlib.pyplot import figure
from matplotlib.path import Path
from matplotlib.patches import PathPatch, Patch
from matplotlib.collections import PatchCollection
from matplotlib.pyplot import rcParams
from py2md.classes import MDHeading, MDTable
from ..results.sectionresult import SectionResult
from .. import config

if TYPE_CHECKING:
    from .material import Material
    from .materialsection import MaterialSection

class CompositeSection():
    sections: List['MaterialSection'] = None
    label: str = None
    _EA: float = None
    _EAy: float = None
    _EAz: float = None
    _cy: float = None
    _cz: float = None
    _EAyy: float = None
    _EAzz: float = None
    _EAyz: float = None
    _EIyy: float = None
    _EIzz: float = None
    _EIyz: float = None
    _thp: float = None
    _EIyp: float = None
    _EIzp: float = None
    def __init__(self, sections: List['MaterialSection'], label: str=None) -> None:
        self.sections = sections
        self.label = label
    def reset(self) -> None:
        self._EA = None
        self._EAy = None
        self._EAz = None
        self._cy = None
        self._cz = None
        self._EAyy = None
        self._EAzz = None
        self._EAyz = None
        self._EIyy = None
        self._EIzz = None
        self._EIyz = None
        self._thp = None
        self._EIyp = None
        self._EIzp = None
    @property
    def EA(self) -> float:
        if self._EA is None:
            self._EA = 0.0
            for section in self.sections:
                self._EA += section.EA
        return self._EA
    @property
    def EAy(self) -> float:
        if self._EAy is None:
            self._EAy = 0.0
            for section in self.sections:
                self._EAy += section.EAy
        return self._EAy
    @property
    def EAz(self) -> float:
        if self._EAz is None:
            self._EAz = 0.0
            for section in self.sections:
                self._EAz += section.EAz
        return self._EAz
    @property
    def cy(self) -> float:
        if self._cy is None:
            self._cy = self.EAy/self.EA
        return self._cy
    @property
    def cz(self) -> float:
        if self._cz is None:
            self._cz = self.EAz/self.EA
        return self._cz
    @property
    def EAyy(self) -> float:
        if self._EAyy is None:
            self._EAyy = 0.0
            for section in self.sections:
                self._EAyy += section.EAyy
        return self._EAyy
    @property
    def EAzz(self) -> float:
        if self._EAzz is None:
            self._EAzz = 0.0
            for section in self.sections:
                self._EAzz += section.EAzz
        return self._EAzz
    @property
    def EAyz(self) -> float:
        if self._EAyz is None:
            self._EAyz = 0.0
            for section in self.sections:
                self._EAyz += section.EAyz
        return self._EAyz
    @property
    def EIyy(self) -> float:
        if self._EIyy is None:
            self._EIyy = self.EAzz-self.EA*self.cz**2
        return self._EIyy
    @property
    def EIzz(self) -> float:
        if self._EIzz is None:
            self._EIzz = self.EAyy-self.EA*self.cy**2
        return self._EIzz
    @property
    def EIyz(self) -> float:
        if self._EIyz is None:
            self._EIyz = self.EAyz-self.EA*self.cy*self.cz
        return self._EIyz
    @property
    def thp(self) -> float:
        if self._thp is None:
            tol = 1e-12
            if abs(2*self.EIyz) < tol:
                self._thp = 0.0
            elif abs(self.EIzz-self.EIyy) < tol:
                self._thp = pi/4
            else:
                self._thp = atan(2*self.EIyz/(self.EIzz-self.EIyy))/2
        return self._thp
    @property
    def EIyp(self) -> float:
        if self._EIyp is None:
            c = cos(self.thp)
            s = sin(self.thp)
            self._EIyp = self.EIyy*c**2+self.EIzz*s**2-2*self.EIyz*c*s
        return self._EIyp
    @property
    def EIzp(self) -> float:
        if self._EIzp is None:
            c = cos(self.thp)
            s = sin(self.thp)
            self._EIzp = self.EIyy*s**2+self.EIzz*c**2+2*self.EIyz*c*s
        return self._EIzp
    def plot(self, ax=None, legloc: str='best'):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
        ax.set_aspect('equal')
        prop_cycle = rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color']
        patches = []
        y, z = [], []
        legel = []
        for i, section in enumerate(self.sections):
            verts = []
            codes = []
            for obj in section.path:
                obj.add_path(verts, codes)
            path = Path(verts, codes)
            patch = PathPatch(path)
            y += section.y
            z += section.z
            patches.append(patch)
            label = ''
            if section.label is not None:
                label = section.label
            legel.append(Patch(facecolor=colors[i], alpha=0.6, label=label))
        p = PatchCollection(patches, alpha=0.6)
        p.set_facecolor(colors)
        ax.add_collection(p)
        ax.set_xlim(min(y), max(y))
        ax.set_ylim(min(z), max(z))
        ax.legend(handles=legel, loc=legloc)
        return ax
    def apply_load(self, loadcase, Fx, My, Mz, limit: bool=False):
        sectresults = []
        for section in self.sections:
            sectresult = SectionResult(section, totalsection=self)
            sectresult.set_load(loadcase, Fx, My, Mz, limit=limit)
            sectresults.append(sectresult)
        return sectresults
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<CompositeSection>'
        else:
            outstr = f'<CompositeSection {self.label:s}>'
        return outstr
    def __str__(self) -> str:
        funit = config.funit
        lunit = config.lunit
        l1frm = config.l1frm
        l2frm = config.l2frm
        l3frm = config.l3frm
        l4frm = config.l4frm
        angfrm = config.angfrm
        eiunit = f'{funit:s}.{lunit:s}<sup>2</sup>'
        if self.label is None:
            head = 'Composite Section Properties'
        else:
            head = f'Composite Section Properties - {self.label:s}'
        heading = MDHeading(head, 3)
        mdstr = str(heading)
        table = MDTable()
        table.add_column(f'EA ({funit:s})', l2frm, data=[self.EA])
        table.add_column(f'EAy ({funit:s}.{lunit:s})', l3frm, data=[self.EAy])
        table.add_column(f'EAz ({funit:s}.{lunit:s})', l3frm, data=[self.EAz])
        table.add_column(f'cy ({lunit:s})', l1frm, data=[self.cy])
        table.add_column(f'cz ({lunit:s})', l1frm, data=[self.cz])
        table.add_column(f'EAyy ({eiunit:s})', l4frm, data=[self.EAyy])
        table.add_column(f'EAzz ({eiunit:s})', l4frm, data=[self.EAzz])
        table.add_column(f'EAyz ({eiunit:s})', l4frm, data=[self.EAyz])
        mdstr += str(table)
        table = MDTable()
        table.add_column(f'EI<sub>yy</sub> ({eiunit:s})', l4frm, data=[self.EIyy])
        table.add_column(f'EI<sub>zz</sub> ({eiunit:s})', l4frm, data=[self.EIzz])
        table.add_column(f'EI<sub>yz</sub> ({eiunit:s})', l4frm, data=[self.EIyz])
        table.add_column('&theta;<sub>p</sub> (&deg;)', angfrm,
                         data=[degrees(self.thp)])
        table.add_column(f'EI<sub>yp</sub> ({eiunit:s})', l4frm, data=[self.EIyp])
        table.add_column(f'EI<sub>zp</sub> ({eiunit:s})', l4frm, data=[self.EIzp])
        mdstr += str(table)
        return mdstr
    def _repr_markdown_(self) -> str:
        return self.__str__()

def normalise_composite_section(compsect: CompositeSection,
                                material: 'Material') -> str:
    lunit = config.lunit
    l1frm = config.l1frm
    l2frm = config.l2frm
    l3frm = config.l3frm
    l4frm = config.l4frm
    angfrm = config.angfrm
    emod = material.E
    # Needs material data added to this output for reference.
    if compsect.label is None:
        head = 'Normalised Composite Section Properties'
    else:
        head = f'Normalised Composite Section Properties - {compsect.label:s}'
    heading = MDHeading(head, 3)
    mdstr = str(heading)
    table = MDTable()
    table.add_column(f'A ({lunit:s}<sup>2</sup>)', l2frm, data=[compsect.EA/emod])
    table.add_column(f'Ay ({lunit:s}<sup>3</sup>)', l3frm, data=[compsect.EAy/emod])
    table.add_column(f'Az ({lunit:s}<sup>3</sup>)', l3frm, data=[compsect.EAz/emod])
    table.add_column(f'cy ({lunit:s})', l1frm, data=[compsect.cy])
    table.add_column(f'cz ({lunit:s})', l1frm, data=[compsect.cz])
    table.add_column(f'Ayy ({lunit:s}<sup>4</sup>)', l4frm, data=[compsect.EAyy/emod])
    table.add_column(f'Azz ({lunit:s}<sup>4</sup>)', l4frm, data=[compsect.EAzz/emod])
    table.add_column(f'Ayz ({lunit:s}<sup>4</sup>)', l4frm, data=[compsect.EAyz/emod])
    mdstr += str(table)
    table = MDTable()
    table.add_column(f'I<sub>yy</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                     data=[compsect.EIyy/emod])
    table.add_column(f'I<sub>zz</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                     data=[compsect.EIzz/emod])
    table.add_column(f'I<sub>yz</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                     data=[compsect.EIyz/emod])
    table.add_column('&theta;<sub>p</sub> (&deg;)', angfrm,
                     data=[degrees(compsect.thp)])
    table.add_column(f'I<sub>yp</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                     data=[compsect.EIyp/emod])
    table.add_column(f'I<sub>zp</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                     data=[compsect.EIzp/emod])
    mdstr += str(table)
    return mdstr



from typing import TYPE_CHECKING, Optional, Union, List
from math import floor
from py2md.classes import MDHeading, MDTable
from .. import config

if TYPE_CHECKING:
    from ..general.materialsection import MaterialSection
    from ..general.compositesection import CompositeSection

class SectionResult():
    loadcase: str = None
    limit: bool = None
    msmode: bool = None
    totalsection: Union['MaterialSection', 'CompositeSection'] = None
    materialsection: 'MaterialSection' = None
    Fx: float = None
    My: float = None
    Mz: float = None
    y: List[float] = None
    z: List[float] = None
    yna: float = None
    zna: float = None
    eps: List[float] = None
    allowed: List[float] = None
    sigma: List[float] = None
    result: List[float] = None
    def __init__(self, materialsection: 'MaterialSection',
                 totalsection: Optional['CompositeSection']=None) -> None:
        self.materialsection = materialsection
        if totalsection is None:
            self.totalsection = materialsection
        else:
            self.totalsection = totalsection
        self.msmode = False
    def set_load(self, loadcase, Fx, My, Mz, limit: bool=False) -> None:
        self.loadcase = loadcase
        self.limit = limit
        self.Fx = Fx
        self.My = My
        self.Mz = Mz
        if limit:
            tensall = self.materialsection.material.Fty
            compall = -self.materialsection.material.Fcy
        else:
            tensall = self.materialsection.material.Ftu
            compall = -self.materialsection.material.Fcu
        Emod = self.materialsection.material.E
        EArea = self.totalsection.EA
        EIyy = self.totalsection.EIyy
        EIzz = self.totalsection.EIzz
        EIyz = self.totalsection.EIyz
        cy = self.totalsection.cy
        cz = self.totalsection.cz
        kd = EIyy*EIzz-EIyz**2
        ky = (Mz*EIyy - My*EIyz)/kd
        kz = (My*EIzz - Mz*EIyz)/kd
        self.y, self.z = [], []
        self.yna, self.zna = [], []
        self.eps = []
        self.allowed = []
        self.sigma = []
        self.result = []
        for point in self.materialsection.pnts:
            self.y.append(point.y)
            self.z.append(point.z)
            yna = point.y-cy
            zna = point.z-cz
            self.yna.append(yna)
            self.zna.append(zna)
            eps = Fx/EArea + ky*yna+kz*zna
            self.eps.append(eps)
            sigma = eps*Emod
            self.sigma.append(sigma)
            allowed = tensall
            result = float('inf')
            if sigma < 0.0:
                allowed = compall
                result = allowed/sigma
            elif sigma > 0.0:
                allowed = tensall
                result = allowed/sigma
            self.allowed.append(allowed)
            self.result.append(result)
    def _repr_markdown_(self) -> str:
        lctyp = 'Ultimate'
        if self.limit:
            lctyp = 'Limit'
        msl = self.materialsection.label
        mdstr = ''
        heading = MDHeading(f'Result of {msl:s} for {self.loadcase:s} ({lctyp:s})', 3)
        mdstr += str(heading)
        table = MDTable()
        table.add_column(f'y ({config.lunit:s})', config.l1frm)
        table.add_column(f'z ({config.lunit:s})', config.l1frm)
        table.add_column(f'y<sub>na</sub> ({config.lunit:s})', config.l1frm)
        table.add_column(f'z<sub>na</sub> ({config.lunit:s})', config.l1frm)
        table.add_column('&epsilon;', '.6f')
        table.add_column(f'&sigma; ({config.sunit:s})', '.1f')
        table.add_column(f'Allowed ({config.sunit:s})', '.1f')
        if config.msmode:
            table.add_column('MS', '.2f')
        else:
            table.add_column('RF', '.2f')
        for i, resi in enumerate(self.result):
            resi = floor(resi*100)/100
            if config.msmode:
                resi = resi-1.0
            table.add_row([self.y[i], self.z[i], self.yna[i], self.zna[i],
                           self.eps[i], self.sigma[i], self.allowed[i], resi])
        mdstr += str(table)
        return mdstr
    def __str__(self) -> str:
        lctyp = 'Ultimate'
        if self.limit:
            lctyp = 'Limit'
        msl = self.materialsection.label
        mdstr = ''
        heading = MDHeading(f'Result of {msl:s} for {self.loadcase:s} ({lctyp:s})', 3)
        mdstr += str(heading)
        table = MDTable()
        table.add_column(f'y ({config.lunit:s})', config.l1frm)
        table.add_column(f'z ({config.lunit:s})', config.l1frm)
        table.add_column(f'y_na ({config.lunit:s})', config.l1frm)
        table.add_column(f'z_na ({config.lunit:s})', config.l1frm)
        table.add_column('Strain', '.6f')
        table.add_column(f'Stress ({config.sunit:s})', '.1f')
        table.add_column(f'Allowed ({config.sunit:s})', '.1f')
        if config.msmode:
            table.add_column('MS', '.2f')
        else:
            table.add_column('RF', '.2f')
        for i, resi in enumerate(self.result):
            resi = floor(resi*100)/100
            if config.msmode:
                resi = resi-1.0
            table.add_row([self.y[i], self.z[i], self.yna[i], self.zna[i],
                           self.eps[i], self.sigma[i], self.allowed[i], resi])
        mdstr += str(table)
        return mdstr
    def __repr__(self) -> str:
        return '<SectionResult>'



from typing import List
from math import pi, cos, sin, atan, degrees, atan2, radians
from matplotlib.pyplot import figure
from matplotlib.patches import Rectangle
from py2md.classes import MDHeading, MDTable
from .. import config

class ThinWalledSection():
    y: List[float] = None
    z: List[float] = None
    t: List[float] = None
    label: str = None
    segs: List['WallSegment'] = None
    _A: float = None
    _Ay: float = None
    _Az: float = None
    _cy: float = None
    _cz: float = None
    _Ayy: float = None
    _Azz: float = None
    _Ayz: float = None
    _Iyy: float = None
    _Izz: float = None
    _Iyz: float = None
    _thp: float = None
    _Iyp: float = None
    _Izp: float = None
    def __init__(self, y: list, z: list, t: list, label: str=None) -> None:
        lent = len(t)
        leny = len(y)
        lenz = len(z)
        if leny != lenz:
            print('The length of y does not equal the length of z.')
            return
        if lent != leny and lent != leny-1:
            print('The length of thickness is not consistant with the geometry.')
            return
        self.y = y
        self.z = z
        self.t = t
        self.label = label
        self.generate_segments()
    def check_area(self, display=True) -> None:
        self._A = None
        if self.A < 0.0:
            if display:
                print('Reversed coordinates.')
            self.y.reverse()
            self.z.reverse()
            self.t.reverse()
            self.generate_segments()
        self._A = None
    def generate_segments(self) -> None:
        lent = len(self.t)
        lenp = len(self.y)
        self.segs = []
        for i in range(lent-1):
            ya = self.y[i]
            za = self.z[i]
            yb = self.y[i+1]
            zb = self.z[i+1]
            ts = self.t[i]
            ws = WallSegment(ya, za, yb, zb, ts)
            self.segs.append(ws)
        if lenp > lent:
            ya = self.y[-2]
            za = self.z[-2]
            yb = self.y[-1]
            zb = self.z[-1]
            ts = self.t[-1]
        else:
            ya = self.y[-1]
            za = self.z[-1]
            yb = self.y[0]
            zb = self.z[0]
            ts = self.t[-1]
        ws = WallSegment(ya, za, yb, zb, ts)
        self.segs.append(ws)
        if lenp != lent:
            self.segs[0].set_free_at_a(True)
            self.segs[-1].set_free_at_b(True)
    def reset(self) -> None:
        self._A = None
        self._Ay = None
        self._Az = None
        self._cy = None
        self._cz = None
        self._Ayy = None
        self._Azz = None
        self._Ayz = None
        self._Iyy = None
        self._Izz = None
        self._Iyz = None
        self._thp = None
        self._Iyp = None
        self._Izp = None
        self.check_area(display=False)
    def mirror_y(self) -> None:
        z = [-zi for zi in self.z]
        self.z = z
        self.reset()
        self.generate_segments()
        self.check_area(display=False)
    def mirror_z(self) -> None:
        y = [-yi for yi in self.y]
        self.y = y
        self.reset()
        self.generate_segments()
        self.check_area(display=False)
    def translate(self, yt: float, zt: float) -> None:
        y = [-yi+yt for yi in self.y]
        self.y = y
        z = [-zi+zt for zi in self.z]
        self.z = z
        self.reset()
        self.generate_segments()
        self.check_area(display=False)
    @property
    def A(self) -> float:
        if self._A is None:
            self._A = 0.0
            for seg in self.segs:
                self._A += seg.A
        return self._A
    @property
    def Ay(self) -> float:
        if self._Ay is None:
            self._Ay = 0.0
            for seg in self.segs:
                self._Ay += seg.Ay
        return self._Ay
    @property
    def Az(self) -> float:
        if self._Az is None:
            self._Az = 0.0
            for seg in self.segs:
                self._Az += seg.Az
        return self._Az
    @property
    def cy(self) -> float:
        if self._cy is None:
            self._cy = self.Ay/self.A
        return self._cy
    @property
    def cz(self) -> float:
        if self._cz is None:
            self._cz = self.Az/self.A
        return self._cz
    @property
    def Ayy(self) -> float:
        if self._Ayy is None:
            self._Ayy = 0.0
            for seg in self.segs:
                self._Ayy += seg.Ayy
        return self._Ayy
    @property
    def Azz(self) -> float:
        if self._Azz is None:
            self._Azz = 0.0
            for seg in self.segs:
                self._Azz += seg.Azz
        return self._Azz
    @property
    def Ayz(self) -> float:
        if self._Ayz is None:
            self._Ayz = 0.0
            for seg in self.segs:
                self._Ayz += seg.Ayz
        return self._Ayz
    @property
    def Iyy(self) -> float:
        if self._Iyy is None:
            self._Iyy = self.Azz - self.A*self.cz**2
        return self._Iyy
    @property
    def Izz(self) -> float:
        if self._Izz is None:
            self._Izz = self.Ayy - self.A*self.cy**2
        return self._Izz
    @property
    def Iyz(self) -> float:
        if self._Iyz is None:
            self._Iyz = self.Ayz - self.A*self.cy*self.cz
        return self._Iyz
    @property
    def thp(self) -> float:
        if self._thp is None:
            tol = 1e-12
            if abs(2*self.Iyz) < tol:
                self._thp = 0.0
            elif abs(self.Izz-self.Iyy) < tol:
                self._thp = pi/4
            else:
                self._thp = atan(2*self.Iyz/(self.Izz-self.Iyy))/2
        return self._thp
    @property
    def Iyp(self) -> float:
        if self._Iyp is None:
            c = cos(self.thp)
            s = sin(self.thp)
            self._Iyp = self.Iyy*c**2+self.Izz*s**2-2*self.Iyz*c*s
        return self._Iyp
    @property
    def Izp(self) -> float:
        if self._Izp is None:
            c = cos(self.thp)
            s = sin(self.thp)
            self._Izp = self.Iyy*s**2+self.Izz*c**2+2*self.Iyz*c*s
        return self._Izp
    def plot(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
        from matplotlib.collections import PatchCollection
        rects = []
        for seg in self.segs:
            rects.append(seg.mpl_rectangle())
        patchcol = PatchCollection(rects)
        ax.add_collection(patchcol)
        ax.set_aspect('equal')
        miny = min(self.y)
        maxy = max(self.y)
        minz = min(self.z)
        maxz = max(self.z)
        maxt = max(self.t)
        ax.set_xlim(miny-maxt/2, maxy+maxt/2)
        ax.set_ylim(minz-maxt/2, maxz+maxt/2)
        return ax
    def _repr_markdown_(self) -> str:
        lunit = config.lunit
        l1frm = config.l1frm
        l2frm = config.l2frm
        l3frm = config.l3frm
        l4frm = config.l4frm
        angfrm = config.angfrm
        if self.label is None:
            head = 'Thin-Walled Section Properties'
        else:
            head = f'Thin-Walled Section Properties - {self.label:s}'
        heading = MDHeading(head, 3)
        mdstr = str(heading)
        table = MDTable()
        table.add_column(f'A ({lunit:s}<sup>2</sup>)', l2frm, data=[self.A])
        table.add_column(f'Ay ({lunit:s}<sup>3</sup>)', l3frm, data=[self.Ay])
        table.add_column(f'Az ({lunit:s}<sup>3</sup>)', l3frm, data=[self.Az])
        table.add_column(f'c<sub>y</sub> ({lunit:s})', l1frm, data=[self.cy])
        table.add_column(f'c<sub>z</sub> ({lunit:s})', l1frm, data=[self.cz])
        table.add_column(f'Ayy ({lunit:s}<sup>4</sup>)', l4frm, data=[self.Ayy])
        table.add_column(f'Azz ({lunit:s}<sup>4</sup>)', l4frm, data=[self.Azz])
        table.add_column(f'Ayz ({lunit:s}<sup>4</sup>)', l4frm, data=[self.Ayz])
        mdstr += table._repr_markdown_()
        table = MDTable()
        table.add_column(f'I<sub>yy</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                         data=[self.Iyy])
        table.add_column(f'I<sub>zz</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                         data=[self.Izz])
        table.add_column(f'I<sub>yz</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                         data=[self.Iyz])
        table.add_column('&theta;<sub>p</sub> (&deg;)', angfrm,
                         data=[degrees(self.thp)])
        table.add_column(f'I<sub>yp</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                         data=[self.Iyp])
        table.add_column(f'I<sub>zp</sub> ({lunit:s}<sup>4</sup>)', l4frm,
                         data=[self.Izp])
        mdstr += table._repr_markdown_()
        return mdstr
    def __str__(self) -> str:
        lunit = config.lunit
        l1frm = config.l1frm
        l2frm = config.l2frm
        l3frm = config.l3frm
        l4frm = config.l4frm
        angfrm = config.angfrm
        if self.label is None:
            head = 'Thin-Walled Section Properties'
        else:
            head = f'Thin-Walled Section Properties - {self.label:s}'
        heading = MDHeading(head, 3)
        outstr = str(heading)
        table = MDTable()
        table.add_column(f'A ({lunit:s}^2)', l2frm, data=[self.A])
        table.add_column(f'Ay ({lunit:s}^3)', l3frm, data=[self.Ay])
        table.add_column(f'Az ({lunit:s}^3)', l3frm, data=[self.Az])
        table.add_column(f'c_y ({lunit:s})', l1frm, data=[self.cy])
        table.add_column(f'c_z ({lunit:s})', l1frm, data=[self.cz])
        table.add_column(f'Ayy ({lunit:s}^4)', l4frm, data=[self.Ayy])
        table.add_column(f'Azz ({lunit:s}^4)', l4frm, data=[self.Azz])
        table.add_column(f'Ayz ({lunit:s}^4)', l4frm, data=[self.Ayz])
        outstr += str(table)
        table = MDTable()
        table.add_column(f'I_yy ({lunit:s}^4)', l4frm,
                         data=[self.Iyy])
        table.add_column(f'I_zz ({lunit:s}^4)', l4frm,
                         data=[self.Izz])
        table.add_column(f'I_yz ({lunit:s}^4)', l4frm,
                         data=[self.Iyz])
        table.add_column('th_p (deg)', angfrm,
                         data=[degrees(self.thp)])
        table.add_column(f'I_yp ({lunit:s}^4)', l4frm,
                         data=[self.Iyp])
        table.add_column(f'I_zp ({lunit:s}^4)', l4frm,
                         data=[self.Izp])
        outstr += str(table)
        return outstr
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<ThinWalledSection>'
        else:
            outstr = f'<ThinWalledSection {self.label:s}>'
        return outstr

class WallSegment():
    ya: float = None
    za: float = None
    yb: float = None
    zb: float = None
    ts: float = None
    fa: float = None
    fb: float = None
    dy: float = None
    dz: float = None
    ls: float = None
    th: float = None
    _A: float = None
    _Ay: float = None
    _Az: float = None
    _cy: float = None
    _cz: float= None
    _Ayy: float = None
    _Azz: float = None
    _Ayz: float = None
    _Iyy: float = None
    _Izz: float = None
    _Iyz: float = None
    _thp: float = None
    _Iyp: float = None
    _Izp: float = None
    def __init__(self, ya: float, za: float, yb: float, zb: float, ts: float) -> None:
        self.ya = ya
        self.za = za
        self.yb = yb
        self.zb = zb
        self.ts = ts
        self.update()
    def update(self) -> None:
        self.fa = False
        self.fb = False
        self.dy = self.yb-self.ya
        self.dz = self.zb-self.za
        self.ls = (self.dy**2+self.dz**2)**0.5
        self.th = degrees(atan2(self.dz, self.dy))
    def set_free_at_a(self, fa) -> None:
        self.fa = fa
    def set_free_at_b(self, fb) -> None:
        self.fb = fb
    def is_oef(self) -> bool:
        if self.fa and not self.fb:
            return True
        if not self.fa and self.fb:
            return True
        return False
    def is_nef(self) -> bool:
        if not self.fa and not self.fb:
            return True
        return False
    @property
    def A(self) -> float:
        if self._A is None:
            self._A = self.ts*self.ls
        return self._A
    @property
    def Ay(self) -> float:
        if self._Ay is None:
            self._Ay = self.A*(self.yb+self.ya)/2
        return self._Ay
    @property
    def Az(self) -> float:
        if self._Az is None:
            self._Az = self.A*(self.zb+self.za)/2
        return self._Az
    @property
    def Ayy(self) -> float:
        if self._Ayy is None:
            self._Ayy = self.A*(self.yb**2+self.yb*self.ya+self.ya**2)/3
        return self._Ayy
    @property
    def Azz(self) -> float:
        if self._Azz is None:
            self._Azz = self.A*(self.zb**2+self.zb*self.za+self.za**2)/3
        return self._Azz
    @property
    def Ayz(self) -> float:
        if self._Ayz is None:
            term1 = self.zb*self.yb + self.za*self.ya
            term2 = self.zb*self.ya + self.za*self.yb
            tmp = (term1 + term2/2)/3
            self._Ayz = self.A*tmp
        return self._Ayz
    def mpl_rectangle(self) -> Rectangle:
        thrad = radians(self.th)
        yo = self.ya+self.ts/2*sin(thrad)
        zo = self.za-self.ts/2*cos(thrad)
        rect = Rectangle((yo, zo), self.ls, self.ts, self.th)
        return rect
    def __repr__(self) -> str:
        return '<WallSegment>'



from typing import List, Union, Tuple

from math import atan, cos, sin, degrees, pi, radians
from matplotlib.pyplot import figure
from py2md.classes import MDHeading, MDTable
from .point import Point
from .line import Line
from .arc import Arc, arc_from_points
from .. import config

class GeneralSection():
    y: List[float] = None
    z: List[float] = None
    r: List[float] = None
    label: str = None
    pnts: List[Point] = None
    path: List[Union[Line, Arc]] = None
    _A: float = None
    _Ay: float = None
    _Az: float = None
    _cy: float = None
    _cz: float = None
    _Ayy: float = None
    _Azz: float = None
    _Ayz: float = None
    _Iyy: float = None
    _Izz: float = None
    _Iyz: float = None
    _thp: float = None
    _Iyp: float = None
    _Izp: float = None
    def __init__(self, y: List[float], z: List[float], r: List[float],
                 label: str=None) -> None:
        newy, newz, newr = cleanup_points(y, z, r)
        self.y = newy
        self.z = newz
        self.r = newr
        if label is not None:
            self.label = label
        self.generate_path()
        self.check_area()
    def check_area(self, display=True) -> None:
        self._A = None
        if self.A < 0.0:
            if display:
                print('Reversed coordinates.')
            self.y.reverse()
            self.z.reverse()
            self.r.reverse()
            self.generate_path()
        self._A = None
    def generate_path(self) -> None:
        numpnt = len(self.r)
        pnts = []
        for i in range(numpnt):
            yi = self.y[i]
            zi = self.z[i]
            pnts.append(Point(yi, zi))
        lines: List[Line] = []
        for i in range(numpnt):
            a = i
            b = i+1
            if b == numpnt:
                b = 0
            pnta = pnts[a]
            pntb = pnts[b]
            line = Line(pnta, pntb)
            lines.append(line)
        arcs: List[Arc] = []
        for i in range(numpnt):
            radius = self.r[i]
            if i == 0:
                a = -1
            else:
                a = i-1
            b = i
            linea = lines[a]
            lineb = lines[b]
            pnta = linea.pnta
            pntb = linea.pntb
            pntc = lineb.pntb
            if radius != 0.0:
                arc = arc_from_points(pnta, pntb, pntc, radius)
            else:
                arc = None
            arcs.append(arc)
        lentol = 1e-12
        self.path = []
        for i in range(numpnt):
            a = i
            b = i+1
            if b == numpnt:
                b = 0
            arca = arcs[a]
            arcb = arcs[b]
            linea = lines[a]
            if arca is None:
                pnta = linea.pnta
            else:
                self.path.append(arca)
                pnta = arca.pntb
            if arcb is None:
                pntb = linea.pntb
            else:
                pntb = arcb.pnta
            line = Line(pnta, pntb)
            if line.length > lentol:
                self.path.append(line)
        self.pnts = []
        for obj in self.path:
            self.pnts.append(obj.pnta)
    def reset(self) -> None:
        self._A = None
        self._Ay = None
        self._Az = None
        self._cy = None
        self._cz = None
        self._Ayy = None
        self._Azz = None
        self._Ayz = None
        self._Iyy = None
        self._Izz = None
        self._Iyz = None
        self._thp = None
        self._Iyp = None
        self._Izp = None
        self.check_area(display=False)
    def mirror_y(self) -> None:
        z = [-zi for zi in self.z]
        self.z = z
        self.reset()
        self.generate_path()
        self.check_area(display=False)
    def mirror_z(self) -> None:
        y = [-yi for yi in self.y]
        self.y = y
        self.reset()
        self.generate_path()
        self.check_area(display=False)
    def translate(self, yt: float, zt: float) -> None:
        y = [-yi+yt for yi in self.y]
        self.y = y
        z = [-zi+zt for zi in self.z]
        self.z = z
        self.reset()
        self.generate_path()
        self.check_area(display=False)
    def rotate(self, θr: float) -> None:
        thrad = radians(θr)
        costh = cos(thrad)
        sinth = sin(thrad)
        y = [yi*costh-zi*sinth for yi, zi in zip(self.y, self.z)]
        z = [zi*costh+yi*sinth for yi, zi in zip(self.y, self.z)]
        self.y = y
        self.z = z
        self.reset()
        self.generate_path()
        self.check_area(display=False)
    @property
    def A(self) -> float:
        if self._A is None:
            self._A = 0.0
            for obj in self.path:
                self._A += obj.A
        return self._A
    @property
    def Ay(self) -> float:
        if self._Ay is None:
            self._Ay = 0.0
            for obj in self.path:
                self._Ay += obj.Ay
        return self._Ay
    @property
    def Az(self) -> float:
        if self._Az is None:
            self._Az = 0.0
            for obj in self.path:
                self._Az += obj.Az
        return self._Az
    @property
    def cy(self) -> float:
        if self._cy is None:
            self._cy = self.Ay/self.A
        return self._cy
    @property
    def cz(self) -> float:
        if self._cz is None:
            self._cz = self.Az/self.A
        return self._cz
    @property
    def Ayy(self) -> float:
        if self._Ayy is None:
            self._Ayy = 0.0
            for obj in self.path:
                self._Ayy += obj.Ayy
        return self._Ayy
    @property
    def Azz(self) -> float:
        if self._Azz is None:
            self._Azz = 0.0
            for obj in self.path:
                self._Azz += obj.Azz
        return self._Azz
    @property
    def Ayz(self) -> float:
        if self._Ayz is None:
            self._Ayz = 0.0
            for obj in self.path:
                self._Ayz += obj.Ayz
        return self._Ayz
    @property
    def Iyy(self) -> float:
        if self._Iyy is None:
            self._Iyy = self.Azz - self.A*self.cz**2
        return self._Iyy
    @property
    def Izz(self) -> float:
        if self._Izz is None:
            self._Izz = self.Ayy - self.A*self.cy**2
        return self._Izz
    @property
    def Iyz(self) -> float:
        if self._Iyz is None:
            self._Iyz = self.Ayz - self.A*self.cy*self.cz
        return self._Iyz
    @property
    def thp(self) -> float:
        if self._thp is None:
            tol = 1e-12
            if abs(2*self.Iyz) < tol:
                self._thp = 0.0
            elif abs(self.Izz-self.Iyy) < tol:
                self._thp = pi/4
            else:
                self._thp = atan(2*self.Iyz/(self.Izz-self.Iyy))/2
        return self._thp
    @property
    def Iyp(self) -> float:
        if self._Iyp is None:
            c = cos(self.thp)
            s = sin(self.thp)
            self._Iyp = self.Iyy*c**2+self.Izz*s**2-2*self.Iyz*c*s
        return self._Iyp
    @property
    def Izp(self) -> float:
        if self._Izp is None:
            c = cos(self.thp)
            s = sin(self.thp)
            self._Izp = self.Iyy*s**2+self.Izz*c**2+2*self.Iyz*c*s
        return self._Izp
    def plot(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
        from matplotlib.path import Path
        from matplotlib.patches import PathPatch
        verts = []
        codes = []
        for obj in self.path:
            obj.add_path(verts, codes)
        path = Path(verts, codes)
        patch = PathPatch(path, alpha=0.8)
        ax.set_aspect('equal')
        ax.add_patch(patch)
        ax.set_xlim(min(self.y), max(self.y))
        ax.set_ylim(min(self.z), max(self.z))
        return ax
    def plot_arc_control(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
        for obj in self.path:
            if isinstance(obj, Arc):
                y = [obj.pntf.y, obj.pnta.y, obj.pntd.y,
                     obj.pnte.y, obj.pntb.y, obj.pntf.y]
                z = [obj.pntf.z, obj.pnta.z, obj.pntd.z,
                     obj.pnte.z, obj.pntb.z, obj.pntf.z]
                ax.plot(y, z)
        return ax
    def build_up_table(self):
        table = MDTable()
        table.add_column('Item', '')
        table.add_column('A', config.l2frm)
        table.add_column('Ay', config.l3frm)
        table.add_column('Az', config.l3frm)
        table.add_column('Ayy', config.l4frm)
        table.add_column('Azz', config.l4frm)
        table.add_column('Ayz', config.l4frm)
        for i, obj in enumerate(self.path):
            table.add_row([i+1, obj.A, obj.Ay, obj.Az, obj.Ayy, obj.Azz, obj.Ayz])
        table.add_row(['Total', self.A, self.Ay, self.Az, self.Ayy, self.Azz, self.Ayz])
        return table
    def section_heading(self, head: str):
        if self.label is None:
            head = f'{head:s}'
        else:
            head = f'{head:s} - {self.label:s}'
        heading = MDHeading(head, 3)
        return str(heading)
    def section_properties(self, nohead: bool=True, outtype: str='md'):
        lunit = config.lunit
        l1frm = config.l1frm
        l2frm = config.l2frm
        l3frm = config.l3frm
        l4frm = config.l4frm
        angfrm = config.angfrm
        mdstr = ''
        if not nohead:
            mdstr += self.section_heading('General Section')
        table = MDTable()
        if outtype == 'md':
            table.add_column(f'A ({lunit:s}<sup>2</sup>)', l2frm, data=[self.A])
            table.add_column(f'Ay ({lunit:s}<sup>3</sup>)', l3frm, data=[self.Ay])
            table.add_column(f'Az ({lunit:s}<sup>3</sup>)', l3frm, data=[self.Az])
            table.add_column(f'c<sub>y</sub> ({lunit:s})', l1frm, data=[self.cy])
            table.add_column(f'c<sub>z</sub> ({lunit:s})', l1frm, data=[self.cz])
            table.add_column(f'Ayy ({lunit:s}<sup>4</sup>)', l4frm, data=[self.Ayy])
            table.add_column(f'Azz ({lunit:s}<sup>4</sup>)', l4frm, data=[self.Azz])
            table.add_column(f'Ayz ({lunit:s}<sup>4</sup>)', l4frm, data=[self.Ayz])
            mdstr += table._repr_markdown_()
        else:
            table.add_column(f'A ({lunit:s}^2)', l2frm, data=[self.A])
            table.add_column(f'Ay ({lunit:s}^3)', l3frm, data=[self.Ay])
            table.add_column(f'Az ({lunit:s}^3)', l3frm, data=[self.Az])
            table.add_column(f'c_y ({lunit:s})', l1frm, data=[self.cy])
            table.add_column(f'c_z ({lunit:s})', l1frm, data=[self.cz])
            table.add_column(f'Ayy ({lunit:s}^4)', l4frm, data=[self.Ayy])
            table.add_column(f'Azz ({lunit:s}^4)', l4frm, data=[self.Azz])
            table.add_column(f'Ayz ({lunit:s}^4)', l4frm, data=[self.Ayz])
            mdstr += str(table)
        table = MDTable()
        if outtype == 'md':
            table.add_column(f'I<sub>yy</sub> ({lunit:s}<sup>4</sup>)',
                            l4frm, data=[self.Iyy])
            table.add_column(f'I<sub>zz</sub> ({lunit:s}<sup>4</sup>)',
                            l4frm, data=[self.Izz])
            table.add_column(f'I<sub>yz</sub> ({lunit:s}<sup>4</sup>)',
                            l4frm, data=[self.Iyz])
            table.add_column('&theta;<sub>p</sub> (&deg;)',
                            angfrm, data=[degrees(self.thp)])
            table.add_column(f'I<sub>yp</sub> ({lunit:s}<sup>4</sup>)',
                            l4frm, data=[self.Iyp])
            table.add_column(f'I<sub>zp</sub> ({lunit:s}<sup>4</sup>)',
                            l4frm, data=[self.Izp])
            mdstr += table._repr_markdown_()
        else:
            table.add_column(f'I_yy ({lunit:s}^4)',
                            l4frm, data=[self.Iyy])
            table.add_column(f'I_zz ({lunit:s}^4)',
                            l4frm, data=[self.Izz])
            table.add_column(f'I_yz ({lunit:s}^4)',
                            l4frm, data=[self.Iyz])
            table.add_column('th_p (deg)',
                            angfrm, data=[degrees(self.thp)])
            table.add_column(f'I_yp ({lunit:s}^4)',
                            l4frm, data=[self.Iyp])
            table.add_column(f'I_zp ({lunit:s}^4)',
                            l4frm, data=[self.Izp])
            mdstr += str(table)
        return mdstr
    def _repr_markdown_(self) -> str:
        return self.section_properties(nohead=False, outtype='md')
    def __str__(self) -> str:
        return self.section_properties(nohead=False, outtype='str')
    def __repr__(self) -> str:
        if self.label is None:
            outstr = '<GeneralSection>'
        else:
            outstr = f'<GeneralSection {self.label:s}>'
        return outstr

def cleanup_points(y: List[float], z: List[float],
                   r: List[float]) -> Tuple[List[float], List[float], List[float]]:
    keep = []
    num = len(y)
    for i in range(num):
        a, b = i-1, i
        if a < 0:
            a = num-1
        ya, za = y[a], z[a]
        yb, zb = y[b], z[b]
        if ya == yb and za == zb:
            keep.append(False)
        else:
            keep.append(True)
    newy, newz, newr = [], [], []
    for i in range(num):
        if keep[i]:
            newy.append(y[i])
            newz.append(z[i])
            newr.append(r[i])
        else:
            print('Duplicate point removed.')
    return newy, newz, newr



from py2md.classes import MDTable
from ..general.generalsection import GeneralSection
from .. import config

class SemiTubeSection(GeneralSection):
    do = None
    di = None
    def __init__(self, do: float, di: float, label: str=None):
        self.do = do
        self.di = di
        ro = self.do/2
        ri = self.di/2
        y = [ro, ro, -ro, -ro, -ri, -ri, ri, ri]
        z = [0.0, ro, ro, 0.0, 0.0, ri, ri, 0.0]
        r = [0.0, ro, ro, 0.0, 0.0, ri, ri, 0.0]
        super().__init__(y, z, r, label=label)
    def __repr__(self):
        if self.label is None:
            outstr = '<Semi-Tube-Section>'
        else:
            outstr = f'<Semi-Tube-Section {self.label:s}>'
        return outstr
    def __str__(self):
        mdstr = self.section_heading('Semi-Tube-Section')
        table = MDTable()
        table.add_column(f'd<sub>o</sub> ({config.lunit:s})', config.l1frm, data=[self.do])
        table.add_column(f'd<sub>i</sub> ({config.lunit:s})', config.l1frm, data=[self.di])
        mdstr += str(table)
        mdstr += self.section_properties()
        return mdstr
    def _repr_markdown_(self):
        return self.__str__()
